/*
 *  Copyright (C) 2002 Alan Hourihane.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 *  Author: Alan Hourihane <alanh@fairlite.demon.co.uk>
 */
#ifdef HAVE_DIX_CONFIG_H
#include <dix-config.h>
#endif


#include "rfb.h"
#include "extnsionst.h"
#define _VNC_SERVER
#include <X11/extensions/vncstr.h>
#ifdef XFree86LOADER
#include "xf86Module.h"
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

DevPrivateKey vncCreateScreenResourcesKey = &vncCreateScreenResourcesKey;
int rfbGCKey;

int VncSelectNotify(ClientPtr client, BOOL onoff);
void VncExtensionInit(void);

static int VncErrorBase;  /* first vnc error number */
static int VncEventBase;  /* first vnc event number */

#define USE_ORIG_RES_CODE 0

#if USE_ORIG_RES_CODE
unsigned long VncResourceGeneration = 0;

static RESTYPE VncNotifyList;

static XID faked;

typedef struct _VncNotifyListRec {
    struct _VncNotifyListRec *next;
    ClientPtr client;
} VncNotifyListRec, *VncNotifyListPtr;
#endif


/**
 * Each X client that uses libVnc to talk to this extension will
 * get recorded by one of these records.
 */
typedef struct _VncClientRec {
    ClientPtr client;
    struct _VncClientRec *next;
    XID fakeID;
    RESTYPE res;
} VncClientRec, *VncClientRecPtr;

static VncClientRecPtr ClientsList = NULL;


/**
 * Remove client record from list
 */
static void
VncRemoveClientRecord(ClientPtr client)
{
   /* XXX need 'deleteresource' flag? */
    VncClientRecPtr p = ClientsList, prev = NULL;
    rfbLog("%s client %p\n", __func__, (void *) client);
    while (p) {
        if (p->client == client) {
            /* remove this one */
            if (prev)
                prev->next = p->next;
            else
                ClientsList = p->next;
            xfree(p);
            return;
        }
        prev = p;
        p = p->next;
    }
}


/**
 * This callback will be called by X's resource manager to delete the
 * given resource.  We create one resource for each client.  When X
 * deletes the resource, we know the client is going away.
 */
static int
VncDestroyClientResourceCallback(pointer pn, XID id)
{
    VncClientRecPtr rec = (VncClientRecPtr) pn;
    VncRemoveClientRecord(rec->client);
    return Success;
}


/**
 * Add a client record to the list of clients (unless already in list).
 * We'll create a new, unique resource for this client.  This is used
 * to detect when an X client goes away.
 */
static void
VncSaveClientRecord(ClientPtr client)
{
    VncClientRecPtr rec = ClientsList;

    rfbLog("%s saving client %p\n", __func__, (void *) client);

    /* look if already in list */
    while (rec) {
        if (rec->client == client) {
            return; /* already in list */
        }
        rec = rec->next;
    }

    /* allocate new client record */
    rec = (VncClientRecPtr) xalloc(sizeof(VncClientRec));
    if (rec) {
        rec->client = client;
        rec->fakeID = FakeClientID(client->index);
        rec->res = CreateNewResourceType(VncDestroyClientResourceCallback);
        if (!AddResource(rec->fakeID, rec->res, rec)) {
            xfree(rec);
        }

        /* insert at head of list */
        rec->next = ClientsList;
        ClientsList = rec;
    }
}



static int
ProcVncQueryVersion(ClientPtr client)
{
    xVncQueryVersionReply 	rep;

    REQUEST_SIZE_MATCH(xVncQueryVersionReq);
    rep.type        	= X_Reply;
    rep.sequenceNumber 	= client->sequence;
    rep.length         	= 0;
    rep.majorVersion  	= VNC_MAJOR_VERSION;
    rep.minorVersion  	= VNC_MINOR_VERSION;
    if(client->swapped)
    {
	register char n;
    	swaps(&rep.sequenceNumber, n);
	swaps(&rep.majorVersion, n);
	swaps(&rep.minorVersion, n);
    }
    (void)WriteToClient(client, SIZEOF(xVncQueryVersionReply),
			(char *)&rep);
    return (client->noClientException);
} /* ProcVncQueryVersion */

static int
ProcVncConnection(ClientPtr client)
{
    REQUEST(xVncConnectionReq);
    xVncConnectionReply 	rep;

    rfbUserAllow( stuff->sock, stuff->accept );

    REQUEST_SIZE_MATCH(xVncConnectionReq);
    rep.type        	= X_Reply;
    rep.sequenceNumber 	= client->sequence;
    rep.length         	= 0;
    rep.sock  		= 0;
    rep.accept  	= 0;
    if(client->swapped)
    {
	register char n;
    	swaps(&rep.sequenceNumber, n);
	swaps(&rep.sock, n);
	swaps(&rep.accept, n);
    }
    (void)WriteToClient(client, SIZEOF(xVncConnectionReply),
			(char *)&rep);
    return (client->noClientException);
} /* ProcVncConnection */

static int
ProcVncSelectNotify(ClientPtr client)
{
    REQUEST(xVncSelectNotifyReq);
    xVncSelectNotifyReply 	rep;

    VncSelectNotify(client, stuff->onoff);

    REQUEST_SIZE_MATCH(xVncSelectNotifyReq);
    rep.type        	= X_Reply;
    rep.sequenceNumber 	= client->sequence;
    rep.length         	= 0;
    if(client->swapped)
    {
	register char n;
    	swaps(&rep.sequenceNumber, n);
    }
    (void)WriteToClient(client, SIZEOF(xVncSelectNotifyReply),
			(char *)&rep);
    return (client->noClientException);

}

static int
ProcVncListConnections(ClientPtr client)
{
    xVncListConnectionsReply 	rep;
    rfbClientPtr cl;
    int count = 0;
    struct in_addr ipaddress;
    xVncConnectionListInfo Info;

    /* count connections */
    for (cl = rfbClientHead; cl; cl = cl->next)
#ifdef CHROMIUM
    /* 
     * Fix this, as it should be generic, but we're only using it
     * for Chromium at the moment, so we only list the connections
     * that have a valid chromium encoding. We should be able to list
     * any type requested. FIXME! XXXX
     * See furthur down this function too!!!
     */
	    if (cl->enableChromiumEncoding)
#endif
	    	count++;

    REQUEST_SIZE_MATCH(xVncListConnectionsReq);
    rep.type        	= X_Reply;
    rep.sequenceNumber 	= client->sequence;
    rep.count		= count;
    rep.length 		= count * sizeof(xVncConnectionListInfo) >> 2;
    if(client->swapped)
    {
	register char n;
    	swaps(&rep.sequenceNumber, n);
    	swaps(&rep.count, n);
    }
    (void)WriteToClient(client, SIZEOF(xVncListConnectionsReply),
			(char *)&rep);

    for (cl = rfbClientHead; cl; cl = cl->next) {
#ifdef CHROMIUM
        /* 
         * Fix this, as it should be generic, but we're only using it
         * for Chromium at the moment, so we only list the connections
         * that have a valid chromium encoding. We should be able to list
         * any type requested. FIXME! XXXX
         */
	if (!cl->enableChromiumEncoding)
	    continue;
#endif
	inet_aton(cl->host, &ipaddress);
	memcpy(&Info, &ipaddress, sizeof(struct in_addr));
	WriteToClient(client, SIZEOF(xVncConnectionListInfo), (char*)&Info);
    }

    return (client->noClientException);
} /* ProcVncListConnections */

#ifdef CHROMIUM
static int
ProcVncChromiumStart(ClientPtr client)
{
    REQUEST(xVncChromiumStartReq);
    xVncChromiumStartReply 	rep;

    rfbSendChromiumStart(stuff->ipaddress, stuff->crServerPort, stuff->mothershipPort);

    REQUEST_SIZE_MATCH(xVncChromiumStartReq);
    rep.type        	= X_Reply;
    rep.sequenceNumber 	= client->sequence;
    rep.length         	= 0;
    if(client->swapped)
    {
	register char n;
    	swaps(&rep.sequenceNumber, n);
    }
    (void)WriteToClient(client, SIZEOF(xVncChromiumStartReply),
			(char *)&rep);
    return (client->noClientException);
} /* ProcVncChromiumStart */

static int
ProcVncChromiumMonitor(ClientPtr client)
{
    REQUEST(xVncChromiumMonitorReq);
    xVncChromiumMonitorReply 	rep;

    rfbChromiumMonitorWindowID(stuff->cr_windowid, stuff->windowid);

    REQUEST_SIZE_MATCH(xVncChromiumMonitorReq);
    rep.type        	= X_Reply;
    rep.sequenceNumber 	= client->sequence;
    rep.length         	= 0;
    if(client->swapped)
    {
	register char n;
    	swaps(&rep.sequenceNumber, n);
    }
    (void)WriteToClient(client, SIZEOF(xVncChromiumMonitorReply),
			(char *)&rep);
    return (client->noClientException);
} /* ProcVncChromiumMonitor */
#endif /* CHROMIUM */

static int
ProcVncDispatch(ClientPtr client)
{
    REQUEST(xReq);

    switch (stuff->data)
    {
	case X_VncQueryVersion:
	    return ProcVncQueryVersion(client);
	case X_VncSelectNotify:
	    return ProcVncSelectNotify(client);
	case X_VncConnection:
	    return ProcVncConnection(client);
	case X_VncListConnections:
	    return ProcVncListConnections(client);
#ifdef CHROMIUM
	case X_VncChromiumStart:
	    return ProcVncChromiumStart(client);
	case X_VncChromiumMonitor:
	    return ProcVncChromiumMonitor(client);
#endif
	default:
	    return BadRequest;
    }
} /* ProcVncDispatch */

static int
SProcVncQueryVersion(ClientPtr client)
{
    REQUEST(xVncQueryVersionReq);
    register char 	n;

    swaps(&stuff->length, n);
    REQUEST_SIZE_MATCH(xVncQueryVersionReq);
    swaps(&stuff->majorVersion, n);
    swaps(&stuff->minorVersion,n);
    return ProcVncQueryVersion(client);
} /* SProcVncQueryVersion */

static int
SProcVncSelectNotify(ClientPtr client)
{
    REQUEST(xVncSelectNotifyReq);
    register char 	n;

    swaps(&stuff->length, n);
    REQUEST_SIZE_MATCH(xVncSelectNotifyReq);
    return ProcVncSelectNotify(client);
} /* SProcVncSelectNotify */

static int
SProcVncListConnections(ClientPtr client)
{
    REQUEST(xVncListConnectionsReq);
    register char 	n;

    swaps(&stuff->length, n);
    REQUEST_SIZE_MATCH(xVncListConnectionsReq);
    return ProcVncListConnections(client);
} /* SProcVncListConnections */

#ifdef CHROMIUM
static int
SProcVncChromiumStart(ClientPtr client)
{
    REQUEST(xVncChromiumStartReq);
    register char 	n;

    swaps(&stuff->ipaddress, n);
    swaps(&stuff->crServerPort, n);
    swaps(&stuff->mothershipPort, n);
    swaps(&stuff->length, n);
    REQUEST_SIZE_MATCH(xVncChromiumStartReq);
    return ProcVncChromiumStart(client);
} /* SProcVncChromiumStart */

static int
SProcVncChromiumMonitor(ClientPtr client)
{
    REQUEST(xVncChromiumMonitorReq);
    register char 	n;

    swaps(&stuff->length, n);
    swaps(&stuff->windowid, n);
    swaps(&stuff->cr_windowid, n);
    REQUEST_SIZE_MATCH(xVncChromiumMonitorReq);
    return ProcVncChromiumMonitor(client);
} /* SProcVncChromiumMonitor */
#endif /* CHROMIUM */

static int
SProcVncConnection(ClientPtr client)
{
    REQUEST(xVncConnectionReq);
    register char 	n;


    swaps(&stuff->length, n);
    REQUEST_SIZE_MATCH(xVncConnectionReq);
    swaps(&stuff->sock, n);
    swaps(&stuff->accept,n);
    return ProcVncConnection(client);
} /* SProcVncConnection */

static int
SProcVncDispatch(ClientPtr client)
{
    REQUEST(xReq);

    switch (stuff->data)
    {
	case X_VncQueryVersion:
	    return SProcVncQueryVersion(client);
	case X_VncSelectNotify:
	    return SProcVncSelectNotify(client);
	case X_VncConnection:
	    return SProcVncConnection(client);
	case X_VncListConnections:
	    return SProcVncListConnections(client);
#ifdef CHROMIUM
	case X_VncChromiumStart:
	    return SProcVncChromiumStart(client);
	case X_VncChromiumMonitor:
	    return SProcVncChromiumMonitor(client);
#endif
	default:
	    return BadRequest;
    }
} /* SProcVncDispatch */

static void 
SwapVncConnectedEvent(xVncConnectedEvent *from, xVncConnectedEvent *to)
{
    to->type = from->type;
    to->detail = from->detail;
    cpswaps(from->sequenceNumber, to->sequenceNumber);
    cpswapl(from->connected, to->connected);
}

#ifdef CHROMIUM
static void 
SwapVncChromiumConnectedEvent(xVncConnectedEvent *from, xVncConnectedEvent *to)
{
    to->type = from->type;
    to->detail = from->detail;
    cpswaps(from->sequenceNumber, to->sequenceNumber);
    cpswapl(from->connected, to->connected);
}
#endif

static void 
SwapVncDisconnectedEvent(xVncConnectedEvent *from, xVncConnectedEvent *to)
{
    to->type = from->type;
    to->detail = from->detail;
    cpswaps(from->sequenceNumber, to->sequenceNumber);
    cpswapl(from->connected, to->connected);
}

int
GenerateVncConnectedEvent(int sock)
{
#if USE_ORIG_RES_CODE
    VncNotifyListPtr pn;

    pn = (VncNotifyListPtr)LookupIDByType(faked, VncNotifyList);
#else
    VncClientRecPtr pn = ClientsList;
#endif

    while (pn)
      {
	if (pn->client) 
	{
    	  xVncConnectedEvent conn;
    	  SOCKLEN_T peer_len;
    	  struct sockaddr_in peer;

    	  conn.type = VncEventBase + XVncConnected;
    	  conn.sequenceNumber = pn->client->sequence;
    	  conn.connected = sock;

    	  peer_len = sizeof(peer);
    	  if (getpeername(sock, (struct sockaddr *) &peer, &peer_len) == -1)
		conn.ipaddress = 0;
   	  else
		conn.ipaddress = (CARD32)peer.sin_addr.s_addr;

	  (void) TryClientEvents(pn->client, NULL, (xEventPtr)&conn, 1,
				 NoEventMask, NoEventMask, NullGrab);
	}
	pn = pn->next;
    }

    return 1;
}



#ifdef CHROMIUM
int
GenerateVncChromiumConnectedEvent(int sock)
{
#if USE_ORIG_RES_CODE
    VncNotifyListPtr pn;
    pn = (VncNotifyListPtr)LookupIDByType(faked, VncNotifyList);
#else
    VncClientRecPtr pn = ClientsList;
#endif
    rfbLog("Enter GenerateVncChromiumConnectedEvent\n");
    while (pn)
    {
	if (pn->client) 
	{
    	  xVncConnectedEvent conn;
    	  SOCKLEN_T peer_len;
    	  struct sockaddr_in peer;

          rfbLog("Sending XVncChromiumConnected to client %p\n",
                 (void *) pn->client);

    	  conn.type = VncEventBase + XVncChromiumConnected;
    	  conn.sequenceNumber = pn->client->sequence;
    	  conn.connected = sock;

    	  peer_len = sizeof(peer);
    	  if (getpeername(sock, (struct sockaddr *)&peer, &peer_len) == -1)
		conn.ipaddress = 0;
   	  else
		conn.ipaddress = (CARD32)peer.sin_addr.s_addr;

	  (void) TryClientEvents(pn->client, NULL, (xEventPtr)&conn, 1,
				 NoEventMask, NoEventMask, NullGrab);
	}
	pn = pn->next;
    }
    return 1;
}
#endif /* CHROMIUM */


int
GenerateVncDisconnectedEvent(int sock)
{
#if USE_ORIG_RES_CODE
    VncNotifyListPtr pn;

    pn = (VncNotifyListPtr)LookupIDByType(faked, VncNotifyList);
#else
    VncClientRecPtr pn = ClientsList;
#endif
    rfbLog("GenerateVncDisconnectedEvent\n");
    while (pn)
    {
	if (pn->client) 
	{
    	  xVncDisconnectedEvent conn;
    	  conn.type = VncEventBase + XVncDisconnected;
    	  conn.sequenceNumber = pn->client->sequence;
    	  conn.connected = sock;
	  (void) TryClientEvents(pn->client, NULL, (xEventPtr)&conn, 1,
				 NoEventMask, NoEventMask, NullGrab);
	}
	pn = pn->next;
    }

    return 1;
}

static void
VncResetProc(ExtensionEntry *extEntry)
{
} /* VncResetProc */



/**
 * When the app calls libVnc's XVncSelectNotify() we wind up here.
 * Either add or remove 'client' from the VncNotifyList depending on 'onoff'.
 */
int
VncSelectNotify(ClientPtr client, BOOL onoff)
{
#if USE_ORIG_RES_CODE
    VncNotifyListPtr head, newRec, tmp, freeRec = NULL;

    rfbLog("%s client %p onoff=%d\n", __func__, (void *) client, (int) onoff);
    if (!faked)
	faked = FakeClientID(client->index);
#else
    /* ignore onoff param */
    VncSaveClientRecord(client);
#endif

#if USE_ORIG_RES_CODE
    /* Get the first VncNotifyListPtr */
    head = (VncNotifyListPtr) LookupIDByType(faked, VncNotifyList);

    /* search list for this client */
    tmp = head;
    while (tmp) {
        if (tmp->client == client) {
            /* found client! */
            if (!onoff)
                tmp->client = NULL;
            return Success;
        }
        if (!tmp->client)
            freeRec = tmp;  /* save ptr to free record */
        tmp = tmp->next;
    }

    /* if we get here, we didn't find client in the list */

    if (!onoff) {
        /* if turning off non-existing client, just return */
        return Success;
    }

    /* OK, add new client to list now */

    if (freeRec) {
        /* re-using a free spot */
        freeRec->client = client;
        return Success;
    }

    /* need to allocate new node */
    if (!(newRec = (VncNotifyListPtr)xalloc(sizeof(VncNotifyListRec))))
        return BadAlloc;
    /* insert at head, just as AddResource() does!!! */
    newRec->next = head;
    newRec->client = client;
    if (!AddResource(faked, VncNotifyList, newRec)) {
        xfree(newRec);
        return BadAlloc;
    }
#endif
    return Success;
}


#if USE_ORIG_RES_CODE
static int
VncDestroyNotifyList(pointer pn, XID id)
{
    VncNotifyListPtr cpn = (VncNotifyListPtr) pn;
    rfbLog("%s client %p\n", __func__, (void *) cpn->client);
    cpn->client = NULL;
    return Success;
}
#endif

static Bool
CreateResourceTypes(void)
{
#if USE_ORIG_RES_CODE
    if (VncResourceGeneration == serverGeneration)
        return TRUE;

    VncResourceGeneration = serverGeneration;

    if (!(VncNotifyList = CreateNewResourceType(VncDestroyNotifyList))) {
        ErrorF("CreateResourceTypes: failed to allocate vnc notify list resource.\n");
        return FALSE;
    }
#endif
    return TRUE;
}



static unsigned long vncExtGeneration = 0;
#if XFREE86VNC
extern Bool VNCInit(ScreenPtr pScreen, unsigned char *FBStart);

/* copied from miscrinit.c */
typedef struct
{
    pointer pbits; /* pointer to framebuffer */
    int width;    /* delta to add to a framebuffer addr to move one row down */
} miScreenInitParmsRec, *miScreenInitParmsPtr;


static Bool
vncCreateScreenResources(ScreenPtr pScreen)
{
    int ret = TRUE;
    CreateScreenResourcesProcPtr CreateScreenResources =
        (CreateScreenResourcesProcPtr)
        dixLookupPrivate(&pScreen->devPrivates, vncCreateScreenResourcesKey);
    miScreenInitParmsPtr pScrInitParms;

    pScrInitParms = (miScreenInitParmsPtr)pScreen->devPrivate;

    if ( pScreen->CreateScreenResources != vncCreateScreenResources ) {
        /* Can't find hook we are hung on */
	xf86DrvMsg(pScreen->myNum, X_WARNING /* X_ERROR */,
                   "vncCreateScreenResources %p called when not in "
                   "pScreen->CreateScreenResources %p n",
		   (void *) vncCreateScreenResources,
                   (void *) pScreen->CreateScreenResources );
    }

    /* Now do our stuff */
    VNCInit(pScreen, pScrInitParms->pbits);

    /* Unhook this function ... */
    pScreen->CreateScreenResources = CreateScreenResources;
    dixSetPrivate(&pScreen->devPrivates, vncCreateScreenResourcesKey, NULL);

    /* ... and call the previous CreateScreenResources fuction, if any */
    if (pScreen->CreateScreenResources) {
        ret = (*pScreen->CreateScreenResources)(pScreen);
    }

#ifdef DEBUG
    ErrorF("vncCreateScreenResources() returns %d\n", ret);
#endif
    return ret;
}
#endif



void
VncExtensionInit(void)
{
    ExtensionEntry	*extEntry;

    if (vncExtGeneration != serverGeneration) {
	unsigned int i;

	vncExtGeneration = serverGeneration;

	// no allocation needed for screen privates
	if (!dixRequestPrivate(&rfbGCKey, sizeof(rfbGCRec)))
	    return;

#if XFREE86VNC
    	for (i = 0 ; i < screenInfo.numScreens; i++) 
	{
      	    dixSetPrivate(&screenInfo.screens[i]->devPrivates,
		    vncCreateScreenResourcesKey,
		    xf86Screens[i]->pScreen->CreateScreenResources);
            xf86Screens[i]->pScreen->CreateScreenResources = vncCreateScreenResources;
    	}
#endif

	gethostname(rfbThisHost, 255);
    }

    if (!CreateResourceTypes())
	    return;

    extEntry = AddExtension(VNC_EXTENSION_NAME,
			    XVncNumberEvents, XVncNumberErrors,
			    ProcVncDispatch, SProcVncDispatch,
                            VncResetProc, StandardMinorOpcode);

    VncErrorBase = extEntry->errorBase;
    VncEventBase = extEntry->eventBase;

    EventSwapVector[VncEventBase + XVncConnected] =
	(EventSwapPtr)SwapVncConnectedEvent;
    EventSwapVector[VncEventBase + XVncDisconnected] =
	(EventSwapPtr)SwapVncDisconnectedEvent;
#ifdef CHROMIUM
    EventSwapVector[VncEventBase + XVncChromiumConnected] =
	(EventSwapPtr)SwapVncChromiumConnectedEvent;
#endif
} /* VncExtensionInit */
